@@ -24,18 +24,21 @@
 # SOFTWARE.
 #
 
+import logging
 from autologging import logged
 from reprlib import repr as _r
+from typing import Literal
 
 from . import Packet, DataFormatType
 from . import (ResponsePacket, SendUnitDataResponsePacket, ReadTagServiceResponsePacket, RegisterSessionResponsePacket,
                UnRegisterSessionResponsePacket, ListIdentityResponsePacket, SendRRDataResponsePacket,
                MultiServiceResponsePacket, ReadTagFragmentedServiceResponsePacket, WriteTagServiceResponsePacket,
-               WriteTagFragmentedServiceResponsePacket, generic_read_response, generic_write_response)
+               WriteTagFragmentedServiceResponsePacket, GenericUnconnectedResponsePacket, GenericConnectedResponsePacket)
 from .. import CommError, RequestError
-from ..bytes_ import pack_uint, pack_udint, pack_dint, print_bytes_msg, pack_usint, PACK_DATA_FUNCTION
-from ..const import (ENCAPSULATION_COMMAND, INSUFFICIENT_PACKETS, DATA_ITEM, ADDRESS_ITEM, EXTENDED_SYMBOL, ELEMENT_TYPE,
-                     TAG_SERVICES_REQUEST, CLASS_CODE, CLASS_TYPE, INSTANCE_TYPE, DATA_TYPE, DATA_TYPE_SIZE, UNCONNECTED_SEND)
+from ..bytes_ import Pack, print_bytes_msg
+from ..const import (EncapsulationCommand, INSUFFICIENT_PACKETS, DataItem, AddressItem, EXTENDED_SYMBOL, ELEMENT_TYPE,
+                     TagService, CLASS_TYPE, INSTANCE_TYPE, DataType, DataTypeSize, ConnectionManagerService,
+                     ClassCode, CommonService, STRUCTURE_READ_REPLY, PRIORITY, TIMEOUT_TICKS, ATTRIBUTE_TYPE)
 
 
 @logged
@@ -45,6 +48,8 @@ class RequestPacket(Packet):
     _timeout = b'\x0a\x00'  # 10
     _encap_command = None
     _response_class = ResponsePacket
+    _response_args = ()
+    _response_kwargs = {}
     type_ = None
     VERBOSE_DEBUG = False
 
@@ -77,18 +82,18 @@ class RequestPacket(Packet):
         try:
             return b''.join([
                 command,
-                pack_uint(length),  # Length UINT
-                pack_dint(self._plc._session),  # Session Handle UDINT
+                Pack.uint(length),  # Length UINT
+                Pack.udint(self._plc._session),  # Session Handle UDINT
                 b'\x00\x00\x00\x00',  # Status UDINT
                 self._plc.attribs['context'],  # Sender Context 8 bytes
-                pack_dint(self._plc.attribs['option']),  # Option UDINT
+                Pack.udint(self._plc.attribs['option']),  # Option UDINT
             ])
 
         except Exception as e:
             raise CommError(e)
 
     def _build_common_packet_format(self, addr_data=None) -> bytes:
-        addr_data = b'\x00\x00' if addr_data is None else pack_uint(len(addr_data)) + addr_data
+        addr_data = b'\x00\x00' if addr_data is None else Pack.uint(len(addr_data)) + addr_data
         msg = self.message
         return b''.join([
             b'\x00\x00\x00\x00',  # Interface Handle: shall be 0 for CIP
@@ -97,7 +102,7 @@ class RequestPacket(Packet):
             self._address_type,
             addr_data,
             self._message_type,
-            pack_uint(len(msg)),
+            Pack.uint(len(msg)),
             msg
         ])
 
@@ -132,9 +137,9 @@ class RequestPacket(Packet):
             self._send(self._build_request())
             self.__log.debug(f'Sent: {self!r}')
             reply = self._receive()
-            response = self._response_class(reply)
+            response = self._response_class(reply, *self._response_args, **self._response_kwargs)
         else:
-            response = self._response_class()
+            response = self._response_class(*self._response_args, **self._response_kwargs)
             response._error = self.error
         self.__log.debug(f'Received: {response!r}')
         return response
@@ -147,14 +152,14 @@ class RequestPacket(Packet):
 
 @logged
 class SendUnitDataRequestPacket(RequestPacket):
-    _message_type = DATA_ITEM['Connected']
-    _address_type = ADDRESS_ITEM['Connection Based']
+    _message_type = DataItem.connected
+    _address_type = AddressItem.connection
     _response_class = SendUnitDataResponsePacket
-    _encap_command = ENCAPSULATION_COMMAND['send_unit_data']
+    _encap_command = EncapsulationCommand.send_unit_data
 
     def __init__(self, plc):
         super().__init__(plc)
-        self._msg = [pack_uint(plc._sequence), ]
+        self._msg = [Pack.uint(plc._sequence), ]
 
 
 @logged
@@ -177,9 +182,9 @@ class ReadTagServiceRequestPacket(SendUnitDataRequestPacket):
             self.error = 'Invalid Tag Request Path'
 
         super().add(
-            bytes([TAG_SERVICES_REQUEST['Read Tag']]),
+            TagService.read_tag,
             request_path,
-            pack_uint(self.elements),
+            Pack.uint(self.elements),
         )
 
     def send(self):
@@ -223,10 +228,10 @@ class ReadTagFragmentedServiceRequestPacket(SendUnitDataRequestPacket):
             offset = 0
             responses = []
             while offset is not None:
-                self._msg.extend([bytes([TAG_SERVICES_REQUEST['Read Tag Fragmented']]),
+                self._msg.extend([TagService.read_tag_fragmented,
                                  self.request_path,
-                                 pack_uint(self.elements),
-                                 pack_dint(offset)])
+                                 Pack.uint(self.elements),
+                                 Pack.dint(offset)])
                 self._send(self._build_request())
                 self.__log.debug(f'Sent: {self!r} (offset={offset})')
                 reply = self._receive()
@@ -235,7 +240,7 @@ class ReadTagFragmentedServiceRequestPacket(SendUnitDataRequestPacket):
                 responses.append(response)
                 if response.service_status == INSUFFICIENT_PACKETS:
                     offset += len(response.bytes_)
-                    self._msg = [pack_uint(self._plc._sequence)]
+                    self._msg = [Pack.uint(self._plc._sequence)]
                 else:
                     offset = None
             if all(responses):
@@ -310,12 +315,10 @@ class WriteTagFragmentedServiceRequestPacket(SendUnitDataRequestPacket):
     def add(self, tag, value, elements=1, tag_info=None):
         try:
             if tag_info['tag_type'] == 'struct':
-                self._packed_type = b'\xA0\x02' + pack_uint(tag_info['data_type']['template']['structure_handle'])
-                dt_size = tag_info['data_type']['template']['structure_size']
+                self._packed_type = STRUCTURE_READ_REPLY + Pack.uint(tag_info['data_type']['template']['structure_handle'])
                 self.data_type = tag_info['data_type']['name']
             else:
-                self._packed_type = pack_uint(DATA_TYPE[self.data_type])
-                dt_size = DATA_TYPE_SIZE[self.data_type]
+                self._packed_type = Pack.uint(DataType[self.data_type])
                 self.data_type = tag_info['data_type']
 
             self.tag = tag
@@ -335,21 +338,21 @@ class WriteTagFragmentedServiceRequestPacket(SendUnitDataRequestPacket):
             segment_size = self._plc.connection_size - (len(self.request_path) + len(self._packed_type)
                                                         + 9)  # 9 = len of other stuff in the path
 
-            pack_func = PACK_DATA_FUNCTION[self.data_type] if self.tag_info['tag_type'] == 'atomic' else lambda x: x
+            pack_func = Pack[self.data_type] if self.tag_info['tag_type'] == 'atomic' else lambda x: x
             segments = (self.value[i:i+segment_size]
                         for i in range(0, len(self.value), segment_size))
 
             offset = 0
-            elements_packed = pack_uint(self.elements)
+            elements_packed = Pack.uint(self.elements)
 
             for i, segment in enumerate(segments, start=1):
                 segment_bytes = b''.join(pack_func(s) for s in segment) if not isinstance(segment, bytes) else segment
                 self._msg.extend((
-                    bytes([TAG_SERVICES_REQUEST["Write Tag Fragmented"]]),
+                    TagService.write_tag_fragmented,
                     self.request_path,
                     self._packed_type,
                     elements_packed,
-                    pack_dint(offset),
+                    Pack.dint(offset),
                     segment_bytes
                 ))
 
@@ -360,7 +363,7 @@ class WriteTagFragmentedServiceRequestPacket(SendUnitDataRequestPacket):
                 self.__log.debug(f'Received: {response!r}')
                 responses.append(response)
                 offset += len(segment_bytes)
-                self._msg = [pack_uint(self._plc._sequence), ]
+                self._msg = [Pack.uint(self._plc._sequence), ]
 
             if all(responses):
                 final_response = responses[-1]
@@ -378,14 +381,14 @@ class MultiServiceRequestPacket(SendUnitDataRequestPacket):
     type_ = 'multi'
     _response_class = MultiServiceResponsePacket
 
-    def __init__(self, plc, sequence=1):
+    def __init__(self, plc):
         super().__init__(plc)
         self.tags = []
         self._msg.extend((
-            bytes([TAG_SERVICES_REQUEST["Multiple Service Packet"]]),  # the Request Service
-            pack_usint(2),  # the Request Path Size length in word
+            CommonService.multiple_service_request,  # the Request Service
+            Pack.usint(2),  # the Request Path Size length in word
             CLASS_TYPE["8-bit"],
-            CLASS_CODE["Message Router"],
+            ClassCode.message_router,
             INSTANCE_TYPE["8-bit"],
             b'\x01',  # Instance 1
         ))
@@ -407,10 +410,10 @@ class MultiServiceRequestPacket(SendUnitDataRequestPacket):
         offset = len(rp_list) * 2 + 2
         offsets = []
         for rp in rp_list:
-            offsets.append(pack_uint(offset))
+            offsets.append(Pack.uint(offset))
             offset += len(rp)
 
-        msg = self._msg + [pack_uint(len(rp_list))] + offsets + rp_list
+        msg = self._msg + [Pack.uint(len(rp_list))] + offsets + rp_list
         return b''.join(msg)
 
     def add_read(self, tag, elements=1, tag_info=None):
@@ -418,7 +421,7 @@ class MultiServiceRequestPacket(SendUnitDataRequestPacket):
         request_path = _create_tag_rp(tag, self._plc.tags, self._plc.use_instance_ids)
         if request_path is not None:
 
-            request_path = bytes([TAG_SERVICES_REQUEST['Read Tag']]) + request_path + pack_uint(elements)
+            request_path = TagService.read_tag + request_path + Pack.uint(elements)
             _tag = {'tag': tag, 'elements': elements, 'tag_info': tag_info, 'rp': request_path, 'service': 'read'}
             message = self.build_message(self.tags + [_tag])
             if len(message) < self._plc.connection_size:
@@ -476,46 +479,44 @@ def _make_write_data_tag(tag_info, value, elements, request_path, fragmented=Fal
     if tag_info['tag_type'] == 'struct':
         if not isinstance(value, bytes):
             raise RequestError('Writing UDTs only supports bytes for value')
-        _dt_value = b'\xA0\x02' + pack_uint(tag_info['data_type']['template']['structure_handle'])
+        _dt_value = b'\xA0\x02' + Pack.uint(tag_info['data_type']['template']['structure_handle'])
         data_type = tag_info['data_type']['name']
 
-    elif data_type not in DATA_TYPE:
+    elif data_type not in DataType:
         raise RequestError("Unsupported data type")
     else:
-        _dt_value = pack_uint(DATA_TYPE[data_type])
+        _dt_value = Pack.uint(DataType[data_type])
 
-    service = bytes([TAG_SERVICES_REQUEST['Write Tag Fragmented' if fragmented else 'Write Tag']])
+    service = TagService.write_tag_fragmented if fragmented else TagService.write_tag
 
     request_path = b''.join((service,
                              request_path,
                              _dt_value,
-                             pack_uint(elements),
+                             Pack.uint(elements),
                              value))
     return request_path, data_type
 
 
 def _make_write_data_bit(tag_info, value, request_path):
-    mask_size = DATA_TYPE_SIZE.get(tag_info['data_type'])
+    mask_size = DataTypeSize.get(tag_info['data_type'])
     if mask_size is None:
         raise RequestError(f'Invalid data type {tag_info["data_type"]} for writing bits')
 
     or_mask, and_mask = value
-    request_path = b''.join((
-        bytes([TAG_SERVICES_REQUEST["Read Modify Write Tag"]]),
-        request_path,
-        pack_uint(mask_size),
-        pack_udint(or_mask)[:mask_size],
-        pack_udint(and_mask)[:mask_size]
-    ))
-
-    return request_path
+    return b''.join((
+            TagService.read_modify_write,
+            request_path,
+            Pack.uint(mask_size),
+            Pack.udint(or_mask)[:mask_size],
+            Pack.udint(and_mask)[:mask_size]
+        ))
 
 
 @logged
 class SendRRDataRequestPacket(RequestPacket):
-    _message_type = DATA_ITEM['Unconnected']
-    _address_type = ADDRESS_ITEM['UCMM']
-    _encap_command = ENCAPSULATION_COMMAND['send_rr_data']
+    _message_type = DataItem.unconnected
+    _address_type = AddressItem.uccm
+    _encap_command = EncapsulationCommand.send_rr_data
     _response_class = SendRRDataResponsePacket
 
     def _build_common_packet_format(self, addr_data=None) -> bytes:
@@ -524,7 +525,7 @@ class SendRRDataRequestPacket(RequestPacket):
 
 @logged
 class RegisterSessionRequestPacket(RequestPacket):
-    _encap_command = ENCAPSULATION_COMMAND['register_session']
+    _encap_command = EncapsulationCommand.register_session
     _response_class = RegisterSessionResponsePacket
 
     def _build_common_packet_format(self, addr_data=None) -> bytes:
@@ -533,7 +534,7 @@ class RegisterSessionRequestPacket(RequestPacket):
 
 @logged
 class UnRegisterSessionRequestPacket(RequestPacket):
-    _encap_command = ENCAPSULATION_COMMAND['unregister_session']
+    _encap_command = EncapsulationCommand.unregister_session
     _response_class = UnRegisterSessionResponsePacket
 
     def _build_common_packet_format(self, addr_data=None) -> bytes:
@@ -545,7 +546,7 @@ class UnRegisterSessionRequestPacket(RequestPacket):
 
 @logged
 class ListIdentityRequestPacket(RequestPacket):
-    _encap_command = ENCAPSULATION_COMMAND['list_identity']
+    _encap_command = EncapsulationCommand.list_identity
     _response_class = ListIdentityResponsePacket
 
     def _build_common_packet_format(self, addr_data=None) -> bytes:
@@ -558,35 +559,35 @@ def _create_tag_rp(tag, tag_cache, use_instance_ids):
     It returns the request packed wrapped around the tag passed.
     If any error it returns none
     """
+
     tags = tag.split('.')
     if tags:
         base, *attrs = tags
-
-        if use_instance_ids and base in tag_cache:
+        base_tag, index = _find_tag_index(base)
+        if use_instance_ids and base_tag in tag_cache:
             rp = [CLASS_TYPE['8-bit'],
-                  CLASS_CODE['Symbol Object'],
+                  ClassCode.symbol_object,
                   INSTANCE_TYPE['16-bit'],
-                  pack_uint(tag_cache[base]['instance_id'])]
+                  Pack.uint(tag_cache[base_tag]['instance_id'])]
         else:
-            base_tag, index = _find_tag_index(base)
             base_len = len(base_tag)
             rp = [EXTENDED_SYMBOL,
-                  pack_usint(base_len),
-                  base_tag]
+                  Pack.usint(base_len),
+                  base_tag.encode()]
             if base_len % 2:
                 rp.append(b'\x00')
-            if index is None:
-                return None
-            else:
-                rp += index
+        if index is None:
+            return None
+        else:
+            rp += _encode_tag_index(index)
 
         for attr in attrs:
             attr, index = _find_tag_index(attr)
             tag_length = len(attr)
             # Create the request path
             attr_path = [EXTENDED_SYMBOL,
-                         pack_usint(tag_length),
-                         attr]
+                         Pack.usint(tag_length),
+                         attr.encode()]
             # Add pad byte because total length of Request path must be word-aligned
             if tag_length % 2:
                 attr_path.append(b'\x00')
@@ -594,14 +595,10 @@ def _create_tag_rp(tag, tag_cache, use_instance_ids):
             if index is None:
                 return None
             else:
-                attr_path += index
+                attr_path += _encode_tag_index(index)
             rp += attr_path
 
-        # At this point the Request Path is completed,
-        request_path = b''.join(rp)
-        request_path = bytes([len(request_path) // 2]) + request_path
-
-        return request_path
+        return Pack.epath(b''.join(rp))
 
     return None
 
@@ -614,7 +611,7 @@ def _find_tag_index(tag):
         tag = t[:t.find('[')]  # Get only the tag part
     else:
         index = []
-    return tag.encode(), _encode_tag_index(index)
+    return tag, index
 
 
 def _encode_tag_index(index):
@@ -622,120 +619,116 @@ def _encode_tag_index(index):
         for idx in index:
             val = int(idx)
             if val <= 0xff:
-                path += [ELEMENT_TYPE["8-bit"], pack_usint(val)]
+                path += [ELEMENT_TYPE["8-bit"], Pack.usint(val)]
             elif val <= 0xffff:
-                path += [ELEMENT_TYPE["16-bit"], pack_uint(val)]
+                path += [ELEMENT_TYPE["16-bit"], Pack.uint(val)]
             elif val <= 0xfffffffff:
-                path += [ELEMENT_TYPE["32-bit"], pack_dint(val)]
+                path += [ELEMENT_TYPE["32-bit"], Pack.dint(val)]
             else:
                 return None  # Cannot create a valid request packet
         return path
 
 
-def generic_read_request(connected=True):
+class GenericConnectedRequestPacket(SendUnitDataRequestPacket):
+    _response_class = GenericConnectedResponsePacket
 
-    base_class = SendUnitDataRequestPacket if connected else SendRRDataRequestPacket
+    def __init__(self, plc):
+        super().__init__(plc)
+        self.service = None
+        self.class_code = None
+        self.instance = None
+        self.attribute = None
+        self.request_data = None
 
-    @logged
-    class GenericReadRequestPacket(base_class):
-        _response_class = generic_read_response(connected)
+    def build(self,
+              service: bytes,
+              class_code: bytes,
+              instance: bytes,
+              attribute: bytes = b'',
+              request_data: bytes = b'',
+              data_format: DataFormatType = None):
 
-        def __init__(self, plc, service: bytes, class_code: bytes, instance: bytes, request_data: bytes = None,
-                     data_format: DataFormatType = None, unconnected_send=False):
-            super().__init__(plc)
-            self.data_format = data_format
-            self.class_code = class_code
-            self.instance = instance
-            self.service = service
-            self.request_data = request_data
-            class_type = CLASS_TYPE.get(len(class_code))
+        self._response_kwargs = {'data_format': data_format}
+        self.class_code = class_code
+        self.instance = instance
+        self.attribute = attribute
+        self.service = service
+        self.request_data = request_data
+        req_path = request_path(class_code, instance, attribute)
 
-            if class_type is None:
-                raise RequestError(f'Invalid Class Code Length ({len(class_code)}), must be 1 or 2 bytes')
-
-            instance_type = INSTANCE_TYPE.get(len(instance))
-            if instance_type is None:
-                raise RequestError(f'Invalid Instance Length ({len(instance)}), must be 1 or 2 bytes')
-
-            if unconnected_send:
-                self.add(_unconnected_send())
-
-            request_path = b''.join((class_type, class_code, instance_type, instance))
-            request_path_len = bytes([len(request_path) // 2])
-            self.add(service, request_path_len, request_path)
-
-            if request_data is not None:
-                self.add(request_data)
-
-        def __repr__(self):
-            return f'{self.__class__.__name__}(service={self.service!r}, class_code={self.class_code!r}, ' \
-                   f'instance={self.instance!r}, request_data={self.request_data!r})'
-
-        def send(self):
-            if not self.error:
-                self._send(self._build_request())
-                self.__log.debug(f'Sent: {self!r}')
-                reply = self._receive()
-                response = self._response_class(reply, data_format=self.data_format)
-            else:
-                response = self._response_class()
-                response._error = self.error
-            self.__log.debug(f'Received: {response!r}')
-            return response
-
-    return GenericReadRequestPacket
+        self.add(service, req_path, request_data)
 
 
-def generic_write_request(connected=True):
-    base_class = SendUnitDataRequestPacket if connected else SendRRDataRequestPacket
+class GenericUnconnectedRequestPacket(SendRRDataRequestPacket):
+    _response_class = GenericUnconnectedResponsePacket
 
-    @logged
-    class GenericWriteRequestPacket(base_class):
-        _response_class = generic_write_response(connected)
+    def __init__(self, plc):
+        super().__init__(plc)
+        self.service = None
+        self.class_code = None
+        self.instance = None
+        self.attribute = None
+        self.request_data = None
 
-        def __init__(self, plc, service: bytes, class_code: bytes, instance: bytes, request_data: bytes = None,
-                     unconnected_send=False):
-            super().__init__(plc)
-            self.class_code = class_code
-            self.instance = instance
-            self.service = service
-            self.request_data = request_data
-            class_type = CLASS_TYPE.get(len(class_code))
+    def build(self,
+              service: bytes,
+              class_code: bytes,
+              instance: bytes,
+              attribute: bytes = b'',
+              request_data: bytes = b'',
+              route_path: bytes = b'',
+              unconnected_send: bool = False,
+              data_format: DataFormatType = None):
+        self._response_kwargs = {'data_format': data_format}
+        self.class_code = class_code
+        self.instance = instance
+        self.attribute = attribute
+        self.service = service
+        self.request_data = request_data
+        req_path = request_path(class_code, instance, attribute)
 
-            if class_type is None:
-                raise RequestError(f'Invalid Class Code Length ({len(class_code)}), must be 1 or 2 bytes')
-
-            instance_type = INSTANCE_TYPE.get(len(instance))
-            if instance_type is None:
-                raise RequestError(f'Invalid Instance Length ({len(instance)}), must be 1 or 2 bytes')
-
-            if unconnected_send:
-                self.add(_unconnected_send())
-
-            request_path = b''.join((class_type, class_code, instance_type, instance))
-            request_path_len = bytes([len(request_path) // 2])
-            self.add(service, request_path_len, request_path)
-
-            if request_data is not None:
-                self.add(request_data)
-
-        def __repr__(self):
-            return f'{self.__class__.__name__}(service={self.service!r}, class_code={self.class_code!r}, ' \
-                   f'instance={self.instance!r}, request_data={self.request_data!r})'
-
-    return GenericWriteRequestPacket
+        if unconnected_send:
+            self.add(wrap_unconnected_send(b''.join((service, req_path, request_data)), route_path))
+        else:
+            self.add(service, req_path, request_data, route_path)
 
 
-def _unconnected_send():
-    return b''.join((
-    UNCONNECTED_SEND,
-    b'\x02',
-    CLASS_TYPE['8-bit'],
-    b'\x06',  # class
-    INSTANCE_TYPE["8-bit"],
-    b'\x01',  # instance
-    b'\x0A',  # priority
-    b'\x0e',  # timeout ticks
-    b'\x06\x00',  # service size
-))
+def wrap_unconnected_send(message, route_path):
+    rp = request_path(class_code=ClassCode.connection_manager, instance=b'\x01')
+    msg_len = len(message)
+    return b''.join(
+        [
+            ConnectionManagerService.unconnected_send,
+            rp,
+            PRIORITY,
+            TIMEOUT_TICKS,
+            Pack.uint(msg_len),
+            message,
+            b'\x00' if msg_len % 2 else b'',
+            route_path
+        ]
+    )
 
+
+def request_path(class_code: bytes, instance: bytes, attribute: bytes = b'', data: bytes = b''):
+    class_type = CLASS_TYPE.get(len(class_code))
+
+    if class_type is None:
+        raise RequestError(f'Invalid Class Code Length ({len(class_code)}), must be 1 or 2 bytes')
+
+    instance_type = INSTANCE_TYPE.get(len(instance))
+    if instance_type is None:
+        raise RequestError(f'Invalid Instance Length ({len(instance)}), must be 1 or 2 bytes')
+
+    path = [class_type, class_code, instance_type, instance]
+
+    if attribute:
+        attr_type = ATTRIBUTE_TYPE.get(len(attribute))
+        if attr_type is None:
+            raise RequestError(f'Invalid Attribute Length ({len(attribute)}), must be 1 or 2 bytes')
+        path += [attr_type, attribute]
+
+    if data:
+        path.append(data)
+
+    return Pack.epath(b''.join(path))
